/* eslint-disable @typescript-eslint/prefer-optional-chain */
import { Tsoa } from '@tsoa/runtime'
import {
  isMethodDeclaration,
  type ClassDeclaration,
  type CallExpression,
  type StringLiteral,
  isDecorator,
} from 'typescript'

import { getDecorators, getDecoratorValues, getProduces, getSecurites } from './decoratorUtils'
import { GenerateMetadataError } from './exceptions'
import { getHeaderType } from './headerTypeHelpers'
import { MetadataGenerator } from './metadataGenerator'
import { MethodGenerator } from './methodGenerator'
import { TypeResolver } from './typeResolver'

export class ControllerGenerator {
  private readonly path?: string
  private readonly tags?: string[]
  private readonly security?: Tsoa.Security[]
  private readonly isHidden?: boolean
  private readonly commonResponses: Array<Tsoa.Response & { status: number }>
  private readonly produces?: string[]

  constructor(
    private readonly node: ClassDeclaration,
    private readonly current: MetadataGenerator,
    private readonly parentSecurity: Tsoa.Security[] = [],
  ) {
    this.path = this.getPath()
    this.tags = this.getTags()
    this.security = this.getSecurity()
    this.isHidden = this.getIsHidden()
    this.commonResponses = this.getCommonResponses()
    this.produces = this.getProduces()
  }

  public IsValid() {
    return !!this.path || this.path === ''
  }

  public Generate(): Tsoa.Controller {
    if (!this.node.parent) {
      throw new GenerateMetadataError("Controller node doesn't have a valid parent source file.")
    }
    if (!this.node.name) {
      throw new GenerateMetadataError("Controller node doesn't have a valid name.")
    }

    const sourceFile = this.node.parent.getSourceFile()

    return {
      location: sourceFile.fileName,
      methods: this.buildMethods(),
      name: this.node.name.text,
      path: this.path || '',
      produces: this.produces,
    }
  }

  private buildMethods() {
    return this.node.members
      .filter(isMethodDeclaration)
      .map(
        (m) =>
          new MethodGenerator(
            m,
            this.current,
            this.commonResponses,
            this.path,
            this.tags,
            this.security,
            this.isHidden,
          ),
      )
      .filter((generator) => generator.IsValid())
      .map((generator) => generator.Generate())
  }

  private getPath() {
    const decorators = getDecorators(this.node, (identifier) => identifier.text === 'SwaggerPath')
    if (!decorators || !decorators.length) {
      return
    }
    if (decorators.length > 1) {
      throw new GenerateMetadataError(`Only one Route decorator allowed in '${this.node.name!.text}' class.`)
    }

    const decorator = decorators[0]
    const expression = decorator.parent as CallExpression
    const decoratorArgument = expression.arguments[0] as StringLiteral
    return decoratorArgument ? `${decoratorArgument.text}` : ''
  }

  //TODO: handle or remove
  private getCommonResponses(): Array<Tsoa.Response & { status: number }> {
    const decorators = getDecorators(this.node, (identifier) => identifier.text === 'Response')
    if (!decorators || !decorators.length) {
      return []
    }

    return decorators.map((decorator) => {
      const expression = decorator.parent as CallExpression

      const [name, description, example] = getDecoratorValues(decorator, this.current.typeChecker)
      if (!name) {
        throw new GenerateMetadataError(`Controller's responses should have an explicit name.`)
      }

      return {
        description: description || '',
        examples: example === undefined ? undefined : [example],
        name,
        status: parseInt(name, 10),
        schema:
          expression.typeArguments && expression.typeArguments.length > 0 && !this.isHidden
            ? new TypeResolver(expression.typeArguments[0], this.current).resolve()
            : undefined,
        headers: getHeaderType(expression.typeArguments, 1, this.current),
      }
    })
  }

  private getTags() {
    const decorators = getDecorators(this.node, (identifier) => identifier.text === 'SwaggerTag')
    if (!decorators || !decorators.length) {
      return
    }
    if (decorators.length > 1) {
      throw new GenerateMetadataError(`Only one Tags decorator allowed in '${this.node.name!.text}' class.`)
    }

    const decorator = decorators[0]
    const expression = decorator.parent as CallExpression

    return expression.arguments.map((a: any) => a.text as string)
  }

  //TODO: handle or remove
  private getSecurity(): Tsoa.Security[] {
    const noSecurityDecorators = getDecorators(this.node, (identifier) => identifier.text === 'NoSecurity')
    const securityDecorators = getDecorators(this.node, (identifier) => identifier.text === 'Security')

    if (noSecurityDecorators?.length && securityDecorators?.length) {
      throw new GenerateMetadataError(
        `NoSecurity decorator cannot be used in conjunction with Security decorator in '${
          this.node.name!.text
        }' class.`,
      )
    }

    if (noSecurityDecorators?.length) {
      return []
    }

    if (!securityDecorators || !securityDecorators.length) {
      return this.parentSecurity
    }

    return securityDecorators.map((d) => getSecurites(d, this.current.typeChecker))
  }

  //TODO: handle or remove
  private getIsHidden(): boolean {
    const hiddenDecorators = getDecorators(this.node, (identifier) => identifier.text === 'Hidden')
    if (!hiddenDecorators || !hiddenDecorators.length) {
      return false
    }
    if (hiddenDecorators.length > 1) {
      throw new GenerateMetadataError(`Only one Hidden decorator allowed in '${this.node.name!.text}' class.`)
    }

    return true
  }

  //TODO: handle or remove
  private getProduces(): string[] | undefined {
    const produces = getProduces(this.node, this.current.typeChecker)
    return produces.length ? produces : undefined
  }
}
